# Calc.py, pygtk calculator, by Reinier Heeres <reinier@heeres.eu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# Change log:
#	2007-04-06: rwh, split mode button into ce/ac
#	2007-04-03: rwh, added scientific/financial mode switching and colors
#	2007-04-02: rwh, increased font size and using Decimal class
#	2007-03-26: rwh, first version

import os
import math
from decimal import Decimal
import decimal

import pygtk
pygtk.require('2.0')
import gtk
import pango

class CalcOp:

	OP_NONE = 0
	OP_ASSIGN = 1
	OP_ADD = 2
	OP_SUB = 3
	OP_MUL = 4
	OP_DIV = 5
	OP_NEG = 6
	OP_SIN = 7
	OP_COS = 8
	OP_TAN = 9
	OP_SQUARE = 10
	OP_SQRT = 11
	OP_POW = 12
	OP_INV = 13
	OP_EXP = 14
	OP_LN = 15
	OP_SINH = 16
	OP_COSH = 17
	OP_TANH = 18
	OP_ADDPCT = 19
	OP_SUBPCT = 20

	ANGLE_RAD = 0
	ANGLE_DEG = 1
	ANGLE_TYPES = 2

	AngleScaling = 1

	OpFunction = {
		OP_NONE: lambda a1, a2: 0.0,
		OP_ASSIGN: lambda a1, a2: a1,
		OP_ASSIGN: lambda a1, a2: a1,
		OP_ADD: lambda a1, a2: a1 + a2,
		OP_SUB: lambda a1, a2: a1 - a2,
		OP_MUL: lambda a1, a2: a1 * a2,
		OP_DIV: lambda a1, a2: a1 / a2,
		OP_NEG: lambda a1, a2: -a1,
		OP_SIN: lambda a1, a2: Decimal(str(math.sin(a1 * CalcOp.AngleScaling))),
		OP_COS: lambda a1, a2: Decimal(str(math.cos(a1 * CalcOp.AngleScaling))),
		OP_TAN: lambda a1, a2: Decimal(str(math.tan(a1 * CalcOp.AngleScaling))),
		OP_SQUARE: lambda a1, a2: a1*a1,
		OP_SQRT: lambda a1, a2: a1.sqrt(),
		OP_POW: lambda a1, a2: Decimal(str(pow(float(float(a1)),float(float(a2))))),
		OP_INV: lambda a1, a2: 1/a1,
		OP_EXP: lambda a1, a2: Decimal(str(math.exp(a1))),
		OP_LN: lambda a1, a2: Decimal(str(math.log(a1))),
		OP_SINH: lambda a1, a2: Decimal(str(math.sinh(a1))),
		OP_COSH: lambda a1, a2: Decimal(str(math.cosh(a1))),
		OP_TANH: lambda a1, a2: Decimal(str(math.cosh(a1))),

		OP_ADDPCT: lambda a1, a2: a1 * (Decimal('1.0') + (a2 / Decimal('100.0'))),
		OP_SUBPCT: lambda a1, a2: a1 * (Decimal('1.0') - (a2 / Decimal('100.0'))),
	}

	def __init__(self, op = OP_NONE, arg1 = None, arg2 = None):
		self.Op = op
		self.Arg1 = arg1
		self.Arg2 = arg2
		self.AngleScaling = 1

	def SetAngleType(self,angle):
		if angle == self.ANGLE_RAD:
			CalcOp.AngleScaling = Decimal('1')
		elif angle == self.ANGLE_DEG:
			CalcOp.AngleScaling = Decimal(str(math.pi / 180))
		else:
			CalcOp.AngleScaling = Decimal('1')

	def Process(self):
		try:
			f = self.OpFunction[self.Op]
			return f(self.Arg1, self.Arg2)
		except:
			print 'CalcOp.Process(): Error, args:'
			print self.Op, self.Arg1, self.Arg2
			return Decimal('0')

class Calc:

	MODE_NORM = 0
	MODE_SCI = 1
	MODE_FIN = 2
	MODES = 3
	MODE_ALL = -1

	FLAG_SPECIAL = 0x01
	FLAG_SMALL = 0x02

	def __init__(self, parent,fontsize=15):
		self.Window = parent
		self.FontSize = fontsize

		self.Window.set_title("Calc")
		self.Window.connect("key_press_event",self.KeypressCB)

		self.CreateDialog()
		self.Window.show_all()
		self.Reset()

		self.Mode = self.MODE_ALL
		self.SetMode(self.MODE_NORM)

		self.AngleType = CalcOp.ANGLE_RAD
		self.SetAngleType(CalcOp.ANGLE_RAD)

		self.ClearBeforeChar = False
		self.RequireOp = False

	def ModifyButtonAppearance(self,button,col,small=False):
		if small:
			width = 35
		else:
			width = 50
		button.get_child().set_size_request(width, 50)
		button.get_child().modify_font(self.FontDesc)
		button.get_child().modify_fg(gtk.STATE_NORMAL,col)

	def CreateButton(self,cap,cb,col,small=False):
		button = gtk.Button(cap)
		self.ModifyButtonAppearance(button,col,small)
		button.connect("clicked", cb)
		button.connect("key_press_event", self.IgnoreKeyCB)
		return button

	def CreateDialog(self):
# Button data: [mode,x,y,caption,callback,flags]
		self.ANGLE_BUTTON = 0
		self.ButtonData = [
			[[self.MODE_SCI], 0, 1, 'Rad', lambda w: self.ToggleAngle(), self.FLAG_SPECIAL],
			[[self.MODE_ALL], 6, 1, 'Mode', lambda w: self.ToggleMode(), self.FLAG_SPECIAL],
#			[[self.MODE_ALL], 12, 1, 'Clear', lambda w: self.Reset(), self.FLAG_SPECIAL],
			[[self.MODE_ALL], 12, 1, 'CE', lambda w: self.ResetInput(), self.FLAG_SPECIAL | self.FLAG_SMALL],
			[[self.MODE_ALL], 13, 1, 'AC', lambda w: self.Reset(), self.FLAG_SPECIAL | self.FLAG_SMALL],

			[[self.MODE_ALL], 6, 2, '7', lambda w: self.AddDigit('7'), 0],
			[[self.MODE_ALL], 8, 2, '8', lambda w: self.AddDigit('8'), 0],
			[[self.MODE_ALL], 10, 2, '9', lambda w: self.AddDigit('9'), 0],

			[[self.MODE_ALL], 6, 3, '4', lambda w: self.AddDigit('4'), 0],
			[[self.MODE_ALL], 8, 3, '5', lambda w: self.AddDigit('5'), 0],
			[[self.MODE_ALL], 10, 3, '6', lambda w: self.AddDigit('6'), 0],

			[[self.MODE_ALL], 6, 4, '1', lambda w: self.AddDigit('1'), 0],
			[[self.MODE_ALL], 8, 4, '2', lambda w: self.AddDigit('2'), 0],
			[[self.MODE_ALL], 10, 4, '3', lambda w: self.AddDigit('3'), 0],

			[[self.MODE_ALL], 6, 5, '0', lambda w: self.AddDigit('0'), 0],
			[[self.MODE_ALL], 8, 5, '.', lambda w: self.AddDot(), 0],
			[[self.MODE_ALL], 10, 5, '=', lambda w: self.Process(pops=100), 0],

			[[self.MODE_ALL], 12, 2, '+', lambda w: self.SetDiadic(CalcOp.OP_ADD), 0],
			[[self.MODE_ALL], 12, 3, '-', lambda w: self.SetDiadic(CalcOp.OP_SUB), 0],
			[[self.MODE_ALL], 12, 4, 'x', lambda w: self.SetDiadic(CalcOp.OP_MUL), 0],
			[[self.MODE_ALL], 12, 5, '/', lambda w: self.SetDiadic(CalcOp.OP_DIV), 0],

			[[self.MODE_SCI], 4, 2, 'x^2', lambda w: self.DoMonadic(CalcOp.OP_SQUARE), 0],
			[[self.MODE_SCI], 4, 3, 'sqrt', lambda w: self.DoMonadic(CalcOp.OP_SQRT), 0],
			[[self.MODE_SCI], 4, 4, '1/x', lambda w: self.DoMonadic(CalcOp.OP_INV), 0],
			[[self.MODE_SCI, self.MODE_FIN], 4, 5, '+/-', lambda w: self.DoMonadic(CalcOp.OP_NEG), 0],

			[[self.MODE_SCI], 2, 2, 'exp', lambda w: self.DoMonadic(CalcOp.OP_EXP), 0],
			[[self.MODE_SCI], 2, 3, 'ln', lambda w: self.DoMonadic(CalcOp.OP_LN), 0],
			[[self.MODE_SCI], 2, 4, 'sinh', lambda w: self.DoMonadic(CalcOp.OP_SINH), 0],
			[[self.MODE_SCI], 2, 5, 'cosh', lambda w: self.DoMonadic(CalcOp.OP_COSH), 0],

			[[self.MODE_SCI], 0, 2, 'sin', lambda w: self.DoMonadic(CalcOp.OP_SIN), 0],
			[[self.MODE_SCI], 0, 3, 'cos', lambda w: self.DoMonadic(CalcOp.OP_COS), 0],
			[[self.MODE_SCI], 0, 4, 'tan', lambda w: self.DoMonadic(CalcOp.OP_TAN), 0],
			[[self.MODE_SCI], 0, 5, 'x^y', lambda w: self.SetDiadic(CalcOp.OP_POW), 0],

			[[self.MODE_SCI], 2, 1, 'pi', lambda w: self.SetConstant(math.pi), 0],
			[[self.MODE_SCI], 4, 1, 'e', lambda w: self.SetConstant(math.e), 0],
			[[self.MODE_ALL], 8, 1, '(', lambda w: self.PushOp(), 0],
			[[self.MODE_ALL], 10, 1, ')', lambda w: self.Process(pops=1), 0],

			[[self.MODE_FIN], 4, 3, '+ %', lambda w: self.SetDiadic(CalcOp.OP_ADDPCT), 0],
			[[self.MODE_FIN], 4, 4, '- %', lambda w: self.SetDiadic(CalcOp.OP_SUBPCT), 0],
		]

		self.ColRed = gtk.gdk.Color(65000,20000,20000)
		self.ColBlue = gtk.gdk.Color(20000,20000,65000)

		self.VContainer = gtk.VBox()
		self.VContainer.set_border_width(100)
		self.FontDesc = pango.FontDescription(str="sans bold %d" % self.FontSize)

		self.Pad = gtk.Table(5, 14, True)
		self.Pad.set_row_spacings(3)
		self.Pad.set_col_spacings(3)

		self.TextBox = gtk.Entry()
		self.TextBox.set_alignment(1.0)
		self.TextBox.connect("key_press_event", self.IgnoreKeyCB)
		self.TextBox.set_size_request(300,50)
		self.TextBox.modify_font(self.FontDesc)
		self.TextBox.set_position(1)
		self.Pad.attach(self.TextBox, 6, 14, 0, 1)

		self.Buttons = []
		for i in range(len(self.ButtonData)):
			modes,x,y,cap,cb,flags = self.ButtonData[i]
			if (flags & self.FLAG_SPECIAL):
				col = self.ColRed
			else:
				col = self.ColBlue
			if (flags & self.FLAG_SMALL):
				xend = x + 1
			else:
				xend = x + 2
#			print 'Button: ',cap,'(x,y),xend',x,y,xend
			button = self.CreateButton(cap,cb,col,small=(flags & self.FLAG_SMALL))
			self.Buttons.append(button)
			self.Pad.attach(button, x, xend, y, y+1)

		self.VContainer.add(self.Pad)
		self.Window.set_canvas(self.VContainer)

	def AddDigit(self, addchar):
		if self.RequireOp:
			self.Reset()
			self.RequireOp = False

		if self.ClearBeforeChar:
			self.TextBox.set_text(addchar)
			self.ClearBeforeChar = False
		else:
			if self.TextBox.get_text() == '0':
				self.TextBox.set_text(addchar)
			else:
				self.TextBox.set_text(self.TextBox.get_text() + addchar)
		return True

	def AddDot(self):
		if self.RequireOp:
			self.Reset()
			self.RequireOp = False

		if self.ClearBeforeChar:
			self.TextBox.set_text('0.')
			self.ClearBeforeChar = False
		else:
			self.TextBox.set_text(self.TextBox.get_text() + '.')

	def RemoveChar(self):
		if self.RequireOp:
			self.Reset()
		else:
			str = self.TextBox.get_text()
			self.TextBox.set_text(str[:-1])
		return True

	def IgnoreKeyCB(self, widget, event):
		return True

	def KeypressCB(self, widget, event):
		key = gtk.gdk.keyval_name(event.keyval)
#		print 'Key: %s' % key
		textkeys = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
		if key in textkeys:
			self.AddDigit(key)
		elif key == 'period':
			self.AddDot()

		keymap = {
			'Return': lambda: self.Process(pops=100),
			'equal': lambda: self.Process(pops=100),
			'plus': lambda: self.SetDiadic(CalcOp.OP_ADD),
			'minus': lambda: self.SetDiadic(CalcOp.OP_SUB),
			'asterisk': lambda: self.SetDiadic(CalcOp.OP_MUL),
			'x': lambda: self.SetDiadic(CalcOp.OP_MUL),
			'slash': lambda: self.SetDiadic(CalcOp.OP_DIV),
			'BackSpace': lambda: self.RemoveChar(),
			'Delete': lambda: self.Reset(),
			'parenleft': self.PushOp,
			'parenright': lambda: self.Process(pops=1)
		}
		if keymap.has_key(key):
			f = keymap[key]
			return f()
		return 0

	def GetNumber(self):
		try:
			return Decimal(self.TextBox.get_text())
		except:
			return Decimal('0')

	def SetNumber(self, val):
		if isinstance(val, float) or isinstance(val, int):
			val = Decimal(str(val))

		if self.Mode == self.MODE_NORM:
			text = str(val)
		elif self.Mode == self.MODE_FIN:
			text = '%.02f' % float(val)
		elif self.Mode == self.MODE_SCI:
			if abs(float(val)) < 0.01 or abs(float(val)) >= 10000:
				text = '%.9e' % float(val)
			else:
				text = '%.9f' % float(val)
		else:
			text = ''
		self.TextBox.set_text(text)

	def SetConstant(self, val):
		self.SetNumber(val)
		self.RequireOp = True

	def SetMode(self, mode):
		if mode is not self.Mode:
			for i in range(len(self.ButtonData)):
				modes,x,y,cap,cb,flags = self.ButtonData[i]
				if self.MODE_ALL in modes or mode in modes:
					self.Buttons[i].show()
				else:
					self.Buttons[i].hide()

		self.Mode = mode

		if self.Mode == self.MODE_NORM:
			decimal.getcontext().prec = 20
		elif self.Mode == self.MODE_SCI:
			decimal.getcontext().prec = 40
		elif self.Mode == self.MODE_FIN:
			decimal.getcontext().prec = 20

	def ToggleMode(self):
		self.SetMode((self.Mode + 1) % self.MODES)

	def SetAngleType(self,a):
		self.AngleType = a
		if self.AngleType == CalcOp.ANGLE_RAD:
			self.Buttons[self.ANGLE_BUTTON].set_label('Rad')
		elif self.AngleType == CalcOp.ANGLE_DEG:
			self.Buttons[self.ANGLE_BUTTON].set_label('Deg')
		self.ModifyButtonAppearance(self.Buttons[self.ANGLE_BUTTON],self.ColRed)

	def ToggleAngle(self):
		self.SetAngleType((self.AngleType +1) % CalcOp.ANGLE_TYPES)

# Set diadic operator. Only process if new number entered
	def SetDiadic(self, op):
		if self.ClearBeforeChar == False:
			self.Process(pops=0)
		self.RequireOp = False
		self.ClearBeforeChar = True
		self.CurOp = CalcOp(op, self.GetNumber())
#		print 'CurOp: ', self.CurOp.Arg1, self.CurOp.Op
		return True

	def DoMonadic(self, op):
		p = CalcOp(op, self.GetNumber())
		p.SetAngleType(self.AngleType) 
		self.SetNumber(p.Process())
		self.RequireOp = True
		return True

	def Process(self,pops=0):
		if self.CurOp == None:
			return False
		self.CurOp.Arg2 = self.GetNumber()
		self.CurOp.SetAngleType(self.AngleType) 
#		print 'Process: ', self.CurOp.Arg1, self.CurOp.Arg2, self.CurOp.Op
		val = self.CurOp.Process()
		self.SetNumber(val)
		if pops > 0:
			self.RequireOp = True
			self.PopOp()
			if pops > 1:
				self.Process(pops=pops-1)
		return True

	def Reset(self):
#		print 'Resetting...'
		self.CurOp = None
		self.OpStack = []
		self.TextBox.set_text('0')
		self.ClearBeforeChar = False
		self.RequireOp = False

	def ResetInput(self):
		self.TextBox.set_text('0')
		
# Put current operation on the stack; doesn't make sense if no operation selected
	def PushOp(self):
		if self.CurOp == None:
			return False
		self.OpStack.append(self.CurOp)
		self.CurOp = None
		self.ClearBeforeChar = True
		return True

# Complete current operation and pop previous operation from stack
	def PopOp(self):
		if len(self.OpStack) == 0:
			self.CurOp = None
			return False
#		self.Process()
		self.CurOp = self.OpStack.pop()
#		self.Process()
		return True

def main():
	win = gtk.Window(gtk.WINDOW_TOPLEVEL)
	t = Calc(win,fontsize=15)
	gtk.main()
	return 0

if __name__ == "__main__":
	main()
